from six.moves.urllib.parse import urlsplit
from base64 import b64encode
import pytest
from flask import Flask
from twilio.request_validator import RequestValidator
from twilio.rest.api.v2010.account.call import CallList
from flask_twilio import Twilio, Response
from twilio.twiml.voice_response import Say


def basic_auth(username, password):
    auth = 'Basic ' + b64encode((username + ':' + password).encode()).decode()
    return {'Authorization': auth}


@pytest.fixture
def twilio():
    app = Flask(__name__)
    app.config['TWILIO_ACCOUNT_SID'] = 'sid'
    app.config['TWILIO_AUTH_TOKEN'] = 'token'
    app.config['TWILIO_FROM'] = '+15005550006'
    twilio = Twilio(app)

    @app.route('/call')
    @twilio.twiml
    def call():
        resp = Response()
        resp.append(Say('Testing, 1, 2, 3.'))
        return resp

    return twilio


@pytest.fixture
def always_valid(monkeypatch):
    monkeypatch.setattr(
        RequestValidator, 'validate', lambda *args, **kargs: True)


@pytest.fixture
def always_invalid(monkeypatch):
    monkeypatch.setattr(
        RequestValidator, 'validate', lambda *args, **kargs: False)


@pytest.fixture
def mock_create_call(monkeypatch):
    ret = dict()

    def store(self, to, from_, url):
        ret['to'] = to
        ret['from_'] = from_
        ret['url'] = url

    monkeypatch.setattr(CallList, 'create', store)
    return ret


def test_get_denied(twilio):
    """Check that GET requests are denied."""
    app = twilio.app
    test_client = app.test_client()
    resp = test_client.get('/call')
    assert resp.status_code == 405


def test_post_denied(twilio):
    """Check that POST requests are denied if they do not carry a valid Twilio
    signature."""
    app = twilio.app
    test_client = app.test_client()
    resp = test_client.post('/call')
    assert resp.status_code == 403


def test_testing_allowed(twilio):
    """Check that POST requests are allowed in testing mode."""
    app = twilio.app
    test_client = app.test_client()
    app.config['TESTING'] = True
    resp = test_client.post('/call')
    assert resp.status_code == 200


def test_valid_allowed(twilio, always_valid):
    """Check that POST requests are allowed if Twilio validates them."""
    app = twilio.app
    test_client = app.test_client()
    resp = test_client.post('/call')
    assert resp.status_code == 200


def test_invalid_allowed(twilio, always_invalid):
    """Check that invalid POST requests are denied if Twilio does not validate
    them."""
    app = twilio.app
    test_client = app.test_client()
    resp = test_client.post('/call')
    assert resp.status_code == 403


def test_no_auth(twilio):
    """Test that server sends auth challenge if secret key is set"""
    app = twilio.app
    test_client = app.test_client()
    app.config['SECRET_KEY'] = 'secret'
    resp = test_client.post('/call')
    assert resp.status_code == 401


def test_wrong_username(twilio):
    """Test that server declines if username is wrong"""
    app = twilio.app
    test_client = app.test_client()
    app.config['SECRET_KEY'] = 'secret'
    resp = test_client.post('/call', headers=basic_auth('bob', 'password'))
    assert resp.status_code == 401


def test_wrong_password(twilio):
    """Test that server declines if username is wrong"""
    app = twilio.app
    test_client = app.test_client()
    app.config['SECRET_KEY'] = 'secret'
    resp = test_client.post('/call', headers=basic_auth('twilio', 'password'))
    assert resp.status_code == 401


def test_auth_correct(twilio, always_valid, mock_create_call):
    """Test that server permits a request that is not spoofed"""
    app = twilio.app
    test_client = app.test_client()
    app.config['SECRET_KEY'] = 'secret'
    with app.test_request_context():
        twilio.call_for('call', to='+15005550006')
    urlparts = urlsplit(mock_create_call['url'])
    username, password = urlparts.netloc.split('@')[0].split(':')
    resp = test_client.post(
        urlparts.path, headers=basic_auth(username, password))
    assert resp.status_code == 200



#!/usr/bin/env python
"""
Demo Flask-Twilio application
"""

import os
import socket

from flask import flash, Flask, render_template, request
from flask_twilio import Twilio, Response
from twilio.base.exceptions import TwilioRestException
from twilio.twiml.messaging_response import Message
from twilio.twiml.voice_response import Say
from jinja2 import DictLoader

# A standard test number.
# See https://www.twilio.com/docs/api/rest/test-credentials.
DEFAULT_NUMBER = '+15005550006'

# Application configuration.
app = Flask(__name__)
app.config['SECRET_KEY'] = os.urandom(24)
app.config['SERVER_NAME'] = socket.getfqdn() + ':5000'
app.config['TWILIO_ACCOUNT_SID'] = os.environ.get('TWILIO_ACCOUNT_SID')
app.config['TWILIO_AUTH_SID'] = os.environ.get('TWILIO_AUTH_SID')
app.config['TWILIO_AUTH_TOKEN'] = os.environ.get('TWILIO_AUTH_TOKEN')
app.config['TWILIO_FROM'] = os.environ.get('TWILIO_FROM', DEFAULT_NUMBER)
twilio = Twilio(app)

# Main form template.
app.jinja_loader = DictLoader({'example.html': '''\
<!doctype html>
<title>Flask-Twilio Test</title>
<meta charset="utf-8">
<meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
<link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/4.0.0/css/bootstrap.min.css" integrity="sha384-Gn5384xqQ1aoWXA+058RXPxPg6fy4IWvTNh0E263XmFcJlSAwiGgFAW/dAiS6JXm" crossorigin="anonymous">
<style>
.vertical-container {
    min-height: 100%;
    min-height: 100vh;
    display: flex;
    align-items: center;
}
</style>
<div class=vertical-container>
    <div class=container>
        <div class=jumbotron>
            <h2>Flask-Twilio Test</h2>
        </div>
        {% with messages = get_flashed_messages(with_categories=true) %}
            {% for category, message in messages %}
                <div class="alert alert-{{ category }}">{{ message }}</div>
            {% endfor %}
        {% endwith %}
        <form method=post>
            <input type=hidden id=say name=say value=1>
            <input type=hidden id=sms name=sms value=1>
            <div class=input-group>
                <input autofocus class=form-control type=tel name=to value="{{ to }}">
                <div class=input-group-append>
                    <button class="btn btn-outline-primary" type=submit>
                        Say + SMS
                    </button>
                    <button type=button class="btn btn-outline-primary dropdown-toggle" data-toggle=dropdown aria-haspopup=true aria-expanded=false>
                    </button>
                    <div class="dropdown-menu dropdown-menu-right">
                        <a class=dropdown-item href="#" onclick="$('form').submit();">Say + SMS</a></li>
                        <a class=dropdown-item href="#" onclick="$('#sms').val(0); $('form').submit();">Say only</a></li>
                        <a class=dropdown-item href="#" onclick="$('#say').val(0); $('form').submit();">SMS only</a></li>
                    </div>
                </div>
            </div>
        </form>
    </div>
</div>
<script src="https://code.jquery.com/jquery-3.2.1.slim.min.js" integrity="sha384-KJ3o2DKtIkvYIK3UENzmM7KCkRr/rE9/Qpg6aAZGJwFDMVNA/GpGFF93hXpG5KkN" crossorigin="anonymous"></script>
<script src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.12.9/umd/popper.min.js" integrity="sha384-ApNbgh9B+Y1QKtv3Rn7W3mgPxhU9K/ScQsAP7hUibX39j7fakFPskvXusvfa0b4Q" crossorigin="anonymous"></script>
<script src="https://maxcdn.bootstrapcdn.com/bootstrap/4.0.0/js/bootstrap.min.js" integrity="sha384-JZR6Spejh4U02d8jOt6vLEHfe/JQGiRRSQQxSfFWpi1MquVdAyjUar5+76PVCmYl" crossorigin="anonymous"></script>
'''})


@app.route('/twiml')
@twilio.twiml
def test_call():
    """View for producing the TwiML document."""
    say = int(request.values.get('say', 1))
    sms = int(request.values.get('sms', 1))
    resp = Response()
    if say:
        resp.append(Say('This is a voice call from Twilio!', voice='female'))
    if sms:
        resp.append(Message('This is an SMS message from Twilio!'))
    return resp


@app.route('/', methods=['GET'])
def index_get():
    """Main form view."""
    return render_template(
        'example.html',
        to=request.values.get('to', DEFAULT_NUMBER))


@app.route('/', methods=['POST'])
def index_post():
    """Main form action."""
    try:
        say = int(request.values.get('say', 1))
        sms = int(request.values.get('sms', 1))
        to = request.values.get('to', None)
        if say:
            twilio.call_for('test_call', to, say=say, sms=sms)
        else:
            twilio.message('This is an SMS message from Twilio!', to)
        flash('Request was successfully sent to Twilio.', 'success')
    except TwilioRestException as e:
        flash('Failed to make call: ' + e.msg, 'danger')
        app.logger.exception('Failed to make call')
    return index_get()


# Start application.
app.run('0.0.0.0')



__version__ = '0.0.6'
__all__ = ('Response', 'Twilio')

from string import ascii_letters, digits
from random import SystemRandom
from functools import wraps
from six.moves.urllib.parse import urlsplit, urlunsplit
from twilio.rest import Client
from twilio.request_validator import RequestValidator
from twilio.twiml import TwiML
from flask import Response as FlaskResponse
from flask import abort, current_app, make_response, request, url_for
from flask import _app_ctx_stack as stack
from itsdangerous import TimestampSigner


rand = SystemRandom()
letters_and_digits = ascii_letters + digits


class Response(FlaskResponse, TwiML):
    """
    A response class for constructing TwiML documents, providing all of
    the verbs that are available through :py:class:`twilio.twiml.Response`.
    See also https://www.twilio.com/docs/api/twiml.
    """

    def __init__(self, *args, **kwargs):
        TwiML.__init__(self)
        FlaskResponse.__init__(self, *args, **kwargs)

    @property
    def response(self):
        return [self.to_xml()]

    @response.setter
    def response(self, value):
        pass


class Twilio(object):
    """This class is used to control Twilio calls."""

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Factory method."""
        app.teardown_appcontext(self.teardown)

    def teardown(self, exception):
        ctx = stack.top
        if hasattr(ctx, 'twilio_client'):
            del ctx.twilio_client
        if hasattr(ctx, 'twilio_validator'):
            del ctx.twilio_validator
        if hasattr(ctx, 'twilio_signer'):
            del ctx.twilio_signer

    @property
    def client(self):
        """
        An application-specific intance of
        :py:class:`twilio.rest.Client`. Primarily for internal use.
        """
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'twilio_client'):
                username = current_app.config.get('TWILIO_AUTH_SID')
                account_sid = current_app.config.get('TWILIO_ACCOUNT_SID')
                if username is None:
                    username = account_sid
                elif account_sid is None:
                    account_sid = username
                password = current_app.config['TWILIO_AUTH_TOKEN']
                ctx.twilio_client = Client(
                    username=username, password=password,
                    account_sid=account_sid)
            return ctx.twilio_client

    @property
    def validator(self):
        """
        An application-specific instance of
        :py:class:`twilio.request_validator.RequestValidator`.
        Primarily for internal use.
        """
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'twilio_validator'):
                ctx.twilio_validator = RequestValidator(
                    current_app.config['TWILIO_AUTH_TOKEN'])
            return ctx.twilio_validator

    @property
    def signer(self):
        """
        An application-specific instance of
        :py:class:`itsdangerous.TimestampSigner`, or ``None`` if no secret key
        is set. Primarily for internal use.
        """
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'twilio_signer'):
                if current_app.secret_key is not None:
                    ctx.twilio_signer = TimestampSigner(
                        current_app.secret_key, 'twilio')
                else:
                    ctx.twilio_signer = None
            return ctx.twilio_signer

    def twiml(self, view_func):
        """Decorator for marking view that will create TwiML documents."""
        @wraps(view_func)
        def wrapper(*args, **kwargs):
            if not(current_app.debug or current_app.testing):
                if request.method != 'POST':
                    abort(405)

                # Perform HTTP Basic authentication if a secret key is set.
                #
                # The username must be `twilio`, and the password must be a
                # validly signed string that was generated less than 10 minutes
                # ago. This guarantees that the Twilio call was initiated by
                # this application, rather than a malicious agent.
                #
                # Note that if we are using HTTP, then a malicious agent can
                # still snoop on the data that we are sending to and from
                # Twilio, and can also spoof our reply to Twilio. Both issues
                # would be addressed by using HTTPS.
                if self.signer is not None:
                    auth = request.authorization
                    authorized = (
                        auth and
                        auth.username == 'twilio' and
                        self.signer.validate(auth.password, max_age=600))
                    if not authorized:
                        # If authorization failed, then issue a challenge.
                        return 'Unauthorized', 401, {
                            'WWW-Authenticate': 'Basic realm="Login Required"'}
                # Validate the Twilio request. This guarantees that the request
                # came from Twilio, rather than some other malicious agent.
                valid = self.validator.validate(
                    request.url,
                    request.form,
                    request.headers.get('X-Twilio-Signature', ''))
                if not valid:
                    # If the request was spoofed, then send '403 Forbidden'.
                    abort(403)
            # Call the view itself.
            rv = view_func(*args, **kwargs)
            # Adjust MIME type and return.
            resp = make_response(rv)
            resp.mimetype = 'text/xml'
            return resp
        wrapper.methods = ('GET', 'POST')
        # Done!
        return wrapper

    def call_for(self, endpoint, to, **values):
        """
        Initiate a Twilio call.

        Parameters
        ----------
        endpoint : `str`
            The view endpoint, as would be passed to :py:func:`flask.url_for`.
        to : `str`
            The destination phone number.
        values : `dict`
            Additional keyword arguments to pass to :py:func:`flask.url_for`.

        Returns
        -------
        call : `twilio.rest.resources.Call`
            An object representing the call in progress.
        """
        # Extract keyword arguments that are intended for `calls.create`
        # instead of `url_for`.
        values = dict(values, _external=True)
        from_ = values.pop('from_', None) or current_app.config['TWILIO_FROM']

        # Construct URL for endpoint.
        url = url_for(endpoint, **values)

        # If we are not in debug or testing mode and a secret key is set, then
        # add HTTP basic auth information to the URL. The username is `twilio`.
        # The password is a random string that has been signed with
        # `itsdangerous`.
        if not(current_app.debug or current_app.testing or self.signer is None):
            urlparts = list(urlsplit(url))
            token = ''.join(rand.choice(letters_and_digits) for i in range(32))
            password = self.signer.sign(token).decode()
            urlparts[1] = 'twilio:' + password + '@' + urlparts[1]
            url = urlunsplit(urlparts)

        # Issue phone call.
        return self.client.calls.create(to=to, from_=from_, url=url)

    def message(self, body, to, **values):
        """
        Send an SMS message with Twilio.

        Parameters
        ----------
        body : `str`
            The body of the text message.
        to : `str`
            The destination phone number.
        values : `dict`
            Additional keyword arguments to pass to
            :py:meth:`twilio.rest.resources.SmsMessages.create`.

        Returns
        -------
        message : :py:class:`twilio.rest.resources.SmsMessage`
            An object representing the message that was sent.
        """
        values = dict(values)
        from_ = values.pop('from_', None) or current_app.config['TWILIO_FROM']
        return self.client.messages.create(
            body=body, to=to, from_=from_, **values)



"""
Flask-Twilio
-------------

Make Twilio voice/SMS calls with Flask
"""
import sys

from setuptools import setup
exec(open('flask_twilio.py').readline())

setup_requires = []
if {'pytest', 'test', 'ptr'}.intersection(sys.argv):
    setup_requires.append('pytest-runner')
if {'build_sphinx'}.intersection(sys.argv):
    setup_requires.extend(['sphinx', 'flask-sphinx-themes'])

setup(
    name='Flask-Twilio',
    version=__version__,
    url='https://github.com/lpsinger/flask-twilio',
    license='BSD',
    author='Leo Singer',
    author_email='leo.singer@ligo.org',
    description='Make Twilio voice/SMS calls with Flask',
    long_description=__doc__,
    py_modules=['flask_twilio'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'itsdangerous',
        'flask>=0.9',
        'six',
        'twilio>=6.0.0'
    ],
    setup_requires=setup_requires,
    tests_require=[
        'pytest',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Communications :: Telephony'
    ]
)
